@@ -29,7 +29,12 @@ except Exception, exc:
     ## fall back to pure python
     protocol = TBinaryProtocol
 
+try:
+    from backports import lzma
+except:
+    lzma = None
 
+import gzip
 import os
 import uuid
 import errno
@@ -39,8 +44,12 @@ import subprocess
 import exceptions
 from cStringIO import StringIO
 
-from .ttypes import StreamItem
-from .ttypes_v0_1_0 import StreamItem as StreamItem_v0_1_0
+from ttypes import StreamItem as StreamItem_v0_3_0
+from ttypes_v0_1_0 import StreamItem as StreamItem_v0_1_0
+from ttypes_v0_2_0 import StreamItem as StreamItem_v0_2_0
+
+class VersionMismatchError(Exception):
+    pass
 
 def serialize(msg):
     '''
@@ -52,11 +61,11 @@ def serialize(msg):
     o_transport.seek(0)
     return o_transport.getvalue()
 
-def deserialize(blob):
+def deserialize(blob, message=StreamItem_v0_3_0):
     '''
     Generate a msg from a serialized binary blob for a single msg
     '''
-    chunk = Chunk(data=blob)
+    chunk = Chunk(data=blob, message=message)
     mesgs = list(chunk)
     assert len(mesgs) == 1, 'got %d messages to deserialize instead of one: %r' % (
         len(mesgs), mesgs)
@@ -94,6 +103,7 @@ class md5_file(object):
     def readAll(self, sz):
         '''
         This method allows TBinaryProtocolAccelerated to actually function.
+
         Copied from here
         http://svn.apache.org/repos/asf/hive/trunk/service/lib/py/thrift/transport/TTransport.py
         '''
@@ -119,9 +129,10 @@ class Chunk(object):
     reader/writer for batches of Thrift messages stored in flat files.
     '''
     def __init__(self, path=None, data=None, file_obj=None, mode='rb',
-                 message=StreamItem):
-        '''
-        Load a chunk from an existing file handle or buffer of data.
+                 message=StreamItem_v0_3_0,
+                 read_wrapper=None, write_wrapper=None,
+        ):
+        '''Load a chunk from an existing file handle or buffer of data.
         If no data is passed in, then chunk starts as empty and
         chunk.add(message) can be called to append to it.
 
@@ -141,8 +152,16 @@ class Chunk(object):
 
         :param data: bytes of data from which to read messages
 
-        :param message: defaults to StreamItem; you can specify your
-        own Thrift-generated class here.
+        :param message: defaults to StreamItem_v0_3_0; you can specify
+        your own Thrift-generated class here.
+
+        :param read_wrapper: a function that takes a deserialized
+        message as input and returns a new object to yield from
+        __iter__
+
+        :param write_wrapper: a function used in Chunk.add(obj) that
+        takes the added object as input and returns another object
+        that is a thrift class that can be serialized.
         '''
         if not fastbinary_import_failure:
             logger.debug('using TBinaryProtocolAccelerated (fastbinary)')
@@ -151,6 +170,9 @@ class Chunk(object):
             logger.warn('import fastbinary failed; falling back to 15x slower TBinaryProtocol: %r'\
                             % fastbinary_import_failure)
 
+        self.read_wrapper = read_wrapper
+        self.write_wrapper = write_wrapper
+
         allowed_modes = ['wb', 'ab', 'rb']
         assert mode in allowed_modes, 'mode=%r not in %r' % (mode, allowed_modes)
         self.mode = mode
@@ -184,14 +206,22 @@ class Chunk(object):
                     exc.errno = errno.EEXIST
                     raise exc
                 if path.endswith('.xz'):
-                    assert mode == 'rb', 'mode=%r for .xz' % mode
-                    ## launch xz child
-                    xz_child = subprocess.Popen(
-                        ['xzcat', path],
-                        stdout=subprocess.PIPE,
-                        stderr=subprocess.PIPE)
-                    file_obj = xz_child.stdout
-                    ## what to do with stderr
+                    if lzma is None:
+                        if mode != 'rb':
+                            raise Exception('backports.lzma is not installed and mode=%r but only "rb" is allowed without backports.lzma' % mode)
+                        ## launch xz child
+                        xz_child = subprocess.Popen(
+                            ['xzcat', path],
+                            stdout=subprocess.PIPE,
+                            stderr=subprocess.PIPE)
+                        file_obj = xz_child.stdout
+                        ## what to do with stderr
+                    else:
+                        file_obj = lzma.open(path, mode)
+                        
+                elif path.endswith('.gz'):
+                    assert mode == 'rb', 'mode=%r for .gz' % mode
+                    file_obj  = gzip.open(path)
                 elif path.endswith('.xz.gpg'):
                     assert mode == 'rb', 'mode=%r for .xz' % mode
                     ## launch xz child
@@ -212,7 +242,14 @@ class Chunk(object):
                 dirname = os.path.dirname(path)
                 if dirname and not os.path.exists(dirname):
                     os.makedirs(dirname)
-                file_obj = open(path, mode)
+                if path.endswith('.gz'):
+                    file_obj = gzip.open(path, mode)
+                elif path.endswith('.xz'):
+                    if lzma is None:
+                        raise Exception('file extension is .xz but backports.lzma is not installed')
+                    file_obj = lzma.open(path, mode)
+                else:
+                    file_obj = open(path, mode)
 
         ## if created without any arguments, then prepare to add
         ## messages to an in-memory file object
@@ -236,8 +273,14 @@ class Chunk(object):
                 raise Exception('mode=%r but specified "data"' % mode)
 
         elif file_obj is not None and hasattr(file_obj, 'mode'):
-            assert file_obj.mode[0] == mode[0], 'file_obj.mode=%r != %r=mode'\
-                % (file_obj.mode, mode)
+            if isinstance(file_obj.mode, int):
+                ## some tools, like python gzip library, use int modes
+                file_obj_mode = {1: 'r', 2: 'w'}[file_obj.mode]
+            else:
+                file_obj_mode = file_obj.mode
+
+            assert file_obj_mode[0] == mode[0], 'file_obj.mode=%r != %r=mode'\
+                % (file_obj_mode, mode)
             ## use the file object for writing out the data as it
             ## happens, i.e. in streaming mode.
 
@@ -260,9 +303,19 @@ class Chunk(object):
     def add(self, msg):
         'add message instance to chunk'
         assert self._o_protocol, 'cannot add to a Chunk instantiated with data'
+        assert self._o_chunk_fh is not None, 'cannot Chunk.add after Chunk.close'
+        if self.write_wrapper is not None:
+            msg = self.write_wrapper(msg)
+        if not (isinstance(msg, self.message) or (type(msg) == self.message)):
+            raise VersionMismatchError(
+                'mismatched type: %s != %s' % (type(msg), self.message))
         msg.write(self._o_protocol)
         self._count += 1
 
+    def flush(self):
+        if self._o_chunk_fh is not None:
+            self._o_transport.flush()
+
     def close(self):
         '''
         Close any chunk file that we might have had open for writing.
@@ -330,25 +383,39 @@ class Chunk(object):
                 ## read it from the thrift protocol instance
                 msg.read(i_protocol)
 
+                if hasattr(msg, 'version'):
+                    ## compare the read version to the default version
+                    ## value on the identified message
+                    if not (msg.version == self.message().version):
+                        raise VersionMismatchError(
+                            'read msg.version = %d != %d = message().version):' % \
+                                (msg.version, self.message().version))
+                
                 ## yield is python primitive for iteration
                 self._count += 1
+
+                if self.read_wrapper is not None:
+                    msg = self.read_wrapper(msg)
                 yield msg
 
             except EOFError:
                 break
 
-def decrypt_and_uncompress(data, gpg_private=None):
+def decrypt_and_uncompress(data, gpg_private=None, tmp_dir='/tmp'):
     '''
     Given a data buffer of bytes, if gpg_key_path is provided, decrypt
     data using gnupg, and uncompress using xz.
-
-    :returns (logs, data): where logs is an array of strings, and data
-    is a binary string.
+    
+    :returns: a tuple of (logs, data), where `logs` is an array of
+      strings and data is a binary string
     '''
     _errors = []
+    tmp_path = os.path.join(tmp_dir, 'tmp-compress-and-encrypt-path-' + uuid.uuid4().hex)
+    if not os.path.exists(tmp_path):
+        os.makedirs(tmp_path)
     if gpg_private is not None:
         ### setup gpg for decryption
-        gpg_dir = '/tmp/%s' % uuid.uuid1()
+        gpg_dir = os.path.join(tmp_path, 'gpg_dir')
         os.makedirs(gpg_dir)
 
         gpg_child = subprocess.Popen(
@@ -379,17 +446,21 @@ def decrypt_and_uncompress(data, gpg_private=None):
         ## remove the gpg_dir
         shutil.rmtree(gpg_dir, ignore_errors=True)
 
-    ## launch xz child
-    xz_child = subprocess.Popen(
-        ['xz', '--decompress'],
-        stdin=subprocess.PIPE,
-        stdout=subprocess.PIPE,
-        stderr=subprocess.PIPE)
-    ## use communicate to pass the data incrementally to the child
-    ## while reading the output, to avoid blocking 
-    data, errors = xz_child.communicate(data)
+    if lzma is not None:
+        data = lzma.decompress(data)
 
-    assert not errors, errors
+    else:
+        ## launch xz child
+        xz_child = subprocess.Popen(
+            ['xz', '--decompress'],
+            stdin=subprocess.PIPE,
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE)
+        ## use communicate to pass the data incrementally to the child
+        ## while reading the output, to avoid blocking 
+        data, errors = xz_child.communicate(data)
+
+        assert not errors, errors
 
     return _errors, data
 
@@ -413,7 +484,7 @@ def compress_and_encrypt(data, gpg_public=None, gpg_recipient='trec-kba'):
 
     if gpg_public is not None:
         ### setup gpg for encryption.  
-        gpg_dir = '/tmp/%s' % uuid.uuid1()
+        gpg_dir = os.path.join(tmp_dir, 'tmp-compress-and-encrypt-' + uuid.uuid1().hex)
         os.makedirs(gpg_dir)
 
         ## Load public key.  Could do this just once, but performance
@@ -447,7 +518,7 @@ def compress_and_encrypt(data, gpg_public=None, gpg_recipient='trec-kba'):
 
     return _errors, data
 
-def compress_and_encrypt_path(path, gpg_public=None, gpg_recipient='trec-kba'):
+def compress_and_encrypt_path(path, gpg_public=None, gpg_recipient='trec-kba', tmp_dir='/tmp'):
     '''
     Given a path in the local file system, compress it using xz, if gpg_public
     is provided, encrypt data using gnupg.
@@ -459,9 +530,13 @@ def compress_and_encrypt_path(path, gpg_public=None, gpg_recipient='trec-kba'):
     assert os.path.exists(path), path
     command = 'xz --compress < ' + path
 
+    tmp_path = os.path.join(tmp_dir, 'tmp-compress-and-encrypt-path-' + uuid.uuid4().hex)
+    if not os.path.exists(tmp_path):
+        os.makedirs(tmp_path)
+
     if gpg_public is not None:
         ### setup gpg for encryption.  
-        gpg_dir = '/tmp/%s' % uuid.uuid1()
+        gpg_dir = os.path.join(tmp_path, 'gpg_dir')
         os.makedirs(gpg_dir)
 
         ## Load public key.  Could do this just once, but performance
@@ -483,8 +558,8 @@ def compress_and_encrypt_path(path, gpg_public=None, gpg_recipient='trec-kba'):
 
     ## we want to capture any errors, so do all the work before
     ## returning.  Store the intermediate result in this temp file:
-    o_path = '/tmp/%s' % uuid.uuid1()
-    e_path = '/tmp/%s' % uuid.uuid1()
+    o_path = os.path.join(tmp_path, 'o_path')
+    e_path = os.path.join(tmp_path, 'e_path')
 
     command += ' 1> ' + o_path 
 
